#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdarg.h>
#include <GL/freeglut.h>
#include "game.h"

#define KEYST_SZ	65536 / 32

void draw_text(float x, float y, float r, float g, float b, const char *fmt, ...);

static void display();
static void idle();
static void reshape(int x, int y);
static void keydown(unsigned char key, int x, int y);
static void keyup(unsigned char key, int x, int y);
static void skeydown(int key, int x, int y);
static void skeyup(int key, int x, int y);
static void mouse(int bn, int st, int x, int y);
static void motion(int x, int y);
static void wheel(int wheel, int dir, int x, int y);

static long prev_time;
static uint32_t keystate[KEYST_SZ];
static bool bnstate[16];
static unsigned int modkeys;

int main(int argc, char **argv)
{
	glutInit(&argc, argv);
	glutInitWindowSize(1024, 600);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_SRGB | GLUT_MULTISAMPLE);

	glutInitContextVersion(3, 0);
	glutInitContextProfile(GLUT_COMPATIBILITY_PROFILE);
	glutInitContextFlags(GLUT_DEBUG);

	glutCreateWindow("ludum dare 42");

	glutDisplayFunc(display);
	glutIdleFunc(idle);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keydown);
	glutKeyboardUpFunc(keyup);
	glutSpecialFunc(skeydown);
	glutSpecialUpFunc(skeyup);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutPassiveMotionFunc(motion);
	glutMouseWheelFunc(wheel);

	if(!game_init()) {
		return 1;
	}
	atexit(game_cleanup);

	prev_time = glutGet(GLUT_ELAPSED_TIME);

	glutMainLoop();
	return 0;
}

void game_quit()
{
	exit(0);
}

bool game_keystate(int key)
{
	int idx = key / 32;
	int bit = key % 32;
	return keystate[idx] & (1 << bit);
}

bool game_bnstate(int bn)
{
	return bnstate[bn];
}

unsigned int game_modkeys()
{
	return modkeys;
}

void draw_text(float x, float y, float r, float g, float b, const char *fmt, ...)
{
	char buf[256], *text = buf;
	va_list ap;

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, win_width, 0, win_height, -1, 1);

	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_LIGHTING);

	glRasterPos2f(1, 1);
	glColor3f(r, g, b);
	while(*text) {
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, *text++);
	}

	glPopAttrib();

	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

static void display()
{
	frame_time = glutGet(GLUT_ELAPSED_TIME);
	frame_dt = (frame_time - prev_time) / 1000.0f;

	game_draw();


	static long frames, fps, prev_fps_upd;
	draw_text(1, 1, 1, 1, 0, "fps: %ld", fps);

	if(frame_time - prev_fps_upd >= 1000) {
		fps = frames;
		frames = 0;
		prev_fps_upd = frame_time;
	}
	frames++;

	glutSwapBuffers();
}

static void idle()
{
	glutPostRedisplay();
}

static void reshape(int x, int y)
{
	glViewport(0, 0, x, y);
	win_width = x;
	win_height = y;
	win_aspect = (float)x / (float)y;

	game_reshape(x, y);
}

static void keydown(unsigned char key, int x, int y)
{
	modkeys = glutGetModifiers();
	keystate[key / 32] |= (1 << (key % 32));
	game_keyboard(key, true);
}

static void keyup(unsigned char key, int x, int y)
{
	modkeys = glutGetModifiers();
	keystate[key / 32] &= ~(1 << (key % 32));
	game_keyboard(key, false);
}

static int conv_skey(int key)
{
	if(key >= GLUT_KEY_F1 && key <= GLUT_KEY_F12) {
		return KEY_F1 + (key - GLUT_KEY_F1);
	}

	switch(key) {
	case GLUT_KEY_LEFT:
		return KEY_LEFT;
	case GLUT_KEY_UP:
		return KEY_UP;
	case GLUT_KEY_RIGHT:
		return KEY_RIGHT;
	case GLUT_KEY_DOWN:
		return KEY_DOWN;
	case GLUT_KEY_PAGE_UP:
		return KEY_PGUP;
	case GLUT_KEY_PAGE_DOWN:
		return KEY_PGDOWN;
	case GLUT_KEY_HOME:
		return KEY_HOME;
	case GLUT_KEY_END:
		return KEY_END;
	case GLUT_KEY_INSERT:
		return KEY_INSERT;
	default:
		break;
	}

	return 0;
}

static void skeydown(int key, int x, int y)
{
	modkeys = glutGetModifiers();
	keystate[key / 32] |= (1 << (key % 32));
	game_keyboard(conv_skey(key), true);
}

static void skeyup(int key, int x, int y)
{
	modkeys = glutGetModifiers();
	keystate[key / 32] &= ~(1 << (key % 32));
	game_keyboard(conv_skey(key), false);
}

static void mouse(int bn, int st, int x, int y)
{
	int idx = bn - GLUT_LEFT_BUTTON;
	bool pressed = st == GLUT_DOWN;

	modkeys = glutGetModifiers();

	if(idx == 3) {
		wheel(0, 1, x, y);
		return;
	} else if(idx == 4) {
		wheel(0, -1, x, y);
		return;
	}

	if(idx < 16) {
		bnstate[idx] = pressed;
	}
	game_mbutton(idx, pressed, x, y);
}

static void motion(int x, int y)
{
	game_mmotion(x, y);
}

static void wheel(int wheel, int dir, int x, int y)
{
	game_mwheel(dir, x, y);
}
